#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2013 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;

namespace CleverComponents.InetSuite
{

	public abstract class DnsRecordPersister
	{
		protected abstract DnsRecord CreateRecord();

		public virtual DnsRecord LoadRecord(string[] parameters)
		{
			DnsRecord result = CreateRecord();
			result.Name = parameters[0];
			result.TTL = StringUtils.StrToInt64Def(parameters[2], 0);
			return result;
		}

		public virtual void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			destination.Add(record.Name);
			destination.Add(DnsUtils.GetRecordTypeStr(record.RecordType));
			destination.Add(record.TTL.ToString());
		}
	}

	public class DnsARecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsARecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsARecord result = (DnsARecord)base.LoadRecord(parameters);
			result.IPAddress = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsARecord)record).IPAddress);
		}
	}

	public class DnsAAAARecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsAAAARecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsAAAARecord result = (DnsAAAARecord)base.LoadRecord(parameters);
			result.IPv6Address = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsAAAARecord)record).IPv6Address);
		}
	}

	public class DnsNSRecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsNSRecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsNSRecord result = (DnsNSRecord)base.LoadRecord(parameters);
			result.NameServer = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsNSRecord)record).NameServer);
		}
	}

	public class DnsCNAMERecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsCNAMERecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsCNAMERecord result = (DnsCNAMERecord)base.LoadRecord(parameters);
			result.PrimaryName = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsCNAMERecord)record).PrimaryName);
		}
	}

	public class DnsSOARecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsSOARecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 10)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsSOARecord result = (DnsSOARecord)base.LoadRecord(parameters);
			result.PrimaryNameServer = parameters[3];
			result.ResponsibleMailbox = parameters[4];
			result.SerialNumber = StringUtils.StrToInt64Def(parameters[5], 0);
			result.RefreshInterval = StringUtils.StrToInt64Def(parameters[6], 0);
			result.RetryInterval = StringUtils.StrToInt64Def(parameters[7], 0);
			result.ExpirationLimit = StringUtils.StrToInt64Def(parameters[8], 0);
			result.MinimumTTL = StringUtils.StrToInt64Def(parameters[9], 0);

			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			DnsSOARecord soaRec = (DnsSOARecord)record;

			destination.Add(soaRec.PrimaryNameServer);
			destination.Add(soaRec.ResponsibleMailbox);
			destination.Add(soaRec.SerialNumber.ToString());
			destination.Add(soaRec.RefreshInterval.ToString());
			destination.Add(soaRec.RetryInterval.ToString());
			destination.Add(soaRec.ExpirationLimit.ToString());
			destination.Add(soaRec.MinimumTTL.ToString());
		}
	}

	public class DnsPTRRecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsPTRRecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsPTRRecord result = (DnsPTRRecord)base.LoadRecord(parameters);
			result.DomainName = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsPTRRecord)record).DomainName);
		}
	}

	public class DnsMXRecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsMXRecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 5)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsMXRecord result = (DnsMXRecord)base.LoadRecord(parameters);
			result.Preference = StringUtils.StrToIntDef(parameters[3], 1);
			result.MailServer = parameters[4];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			DnsMXRecord mxRec = (DnsMXRecord)record;
			destination.Add(mxRec.Preference.ToString());
			destination.Add(mxRec.MailServer);
		}
	}

	public class DnsTXTRecordPersister : DnsRecordPersister
	{
		protected override DnsRecord CreateRecord()
		{
			return new DnsTXTRecord();
		}

		public override DnsRecord LoadRecord(string[] parameters)
		{
			if (parameters.Length < 4)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsTXTRecord result = (DnsTXTRecord)base.LoadRecord(parameters);
			result.Text = parameters[3];
			return result;
		}

		public override void SaveRecord(DnsRecord record, StringCollectionEx destination)
		{
			base.SaveRecord(record, destination);

			destination.Add(((DnsTXTRecord)record).Text);
		}
	}

	public class DnsRecordPersisterFactory
	{
		protected virtual DnsRecordPersister CreatePersister(int recType)
		{
			if ((int)DnsRecordType.A == recType)
				return new DnsARecordPersister();
			if ((int)DnsRecordType.NS == recType)
				return new DnsNSRecordPersister();
			if ((int)DnsRecordType.CNAME == recType)
				return new DnsCNAMERecordPersister();
			if ((int)DnsRecordType.SOA == recType)
				return new DnsSOARecordPersister();
			if ((int)DnsRecordType.PTR == recType)
				return new DnsPTRRecordPersister();
			if ((int)DnsRecordType.MX == recType)
				return new DnsMXRecordPersister();
			if ((int)DnsRecordType.TXT == recType)
				return new DnsTXTRecordPersister();
			if ((int)DnsRecordType.AAAA == recType)
				return new DnsAAAARecordPersister();

			throw new DnsServerError(DnsError.DnsServerFailureCode);
		}

		protected virtual DnsRecordPersister CreatePersister(DnsRecord record)
		{
			if (record is DnsARecord)
				return new DnsARecordPersister();
			if (record is DnsNSRecord)
				return new DnsNSRecordPersister();
			if (record is DnsCNAMERecord)
				return new DnsCNAMERecordPersister();
			if (record is DnsSOARecord)
				return new DnsSOARecordPersister();
			if (record is DnsPTRRecord)
				return new DnsPTRRecordPersister();
			if (record is DnsMXRecord)
				return new DnsMXRecordPersister();
			if (record is DnsTXTRecord)
				return new DnsTXTRecordPersister();
			if (record is DnsAAAARecord)
				return new DnsAAAARecordPersister();

			throw new DnsServerError(DnsError.DnsServerFailureCode);
		}

		public DnsRecord LoadRecord(string source)
		{
			string[] parameters = StringUtils.ExtractQuotedWords(source);

			if (parameters.Length < 3)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsRecordPersister persister = CreatePersister(DnsUtils.GetRecordTypeInt(StringUtils.UpperCase(parameters[1])));
			return persister.LoadRecord(parameters);
		}

		public string SaveRecord(DnsRecord record)
		{
			DnsRecordPersister persister = CreatePersister(record);

			StringCollectionEx list = new StringCollectionEx();
			persister.SaveRecord(record, list);

			return StringUtils.GetQuotedWordsString(list);
		}
	}

	public class DnsCachePersister
	{
		void LoadRecords(DnsRecordList records, StringCollectionEx source, ref int index)
		{
			if (index >= source.Count)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			int count = StringUtils.StrToIntDef(source[index], 0);
			index++;

			if (index + count > source.Count)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsRecordPersisterFactory factory = new DnsRecordPersisterFactory();
			while (count > 0)
			{
				DnsRecord record = factory.LoadRecord(source[index]);
				records.Add(record);
				index++;
				count--;
			}
		}

		void SaveRecords(DnsRecordList records, StringCollectionEx destination)
		{
			destination.Add(records.Count.ToString());

			DnsRecordPersisterFactory factory = new DnsRecordPersisterFactory();
			foreach (DnsRecord rec in records)
			{
				string s = factory.SaveRecord(rec);
				destination.Add(s);
			}
		}

		public virtual DnsCacheEntry Load(StringCollectionEx source)
		{
			if (source.Count < 2)
			{
				throw new DnsServerError(DnsError.DnsServerFailureCode);
			}

			DnsCacheEntry cache = new DnsCacheEntry();

			int index = 0;
			cache.RecordType = DnsUtils.GetRecordTypeInt(source[index]);
			index++;

			cache.Name = source[index];
			index++;

			LoadRecords(cache.Answers, source, ref index);
			LoadRecords(cache.NameServers, source, ref index);
			LoadRecords(cache.AdditionalRecords, source, ref index);

			return cache;
		}

		public virtual void Save(DnsCacheEntry cache, StringCollectionEx destination)
		{
			destination.Clear();
			destination.Add(DnsUtils.GetRecordTypeStr(cache.RecordType));
			destination.Add(cache.Name);

			SaveRecords(cache.Answers, destination);
			SaveRecords(cache.NameServers, destination);
			SaveRecords(cache.AdditionalRecords, destination);
		}
	}

	public class DnsZoneManager
	{
		public const string DnzZoneFileExt = ".txt";

		string zonePath;
		DnsRecordList records;

		public DnsZoneManager()
		{
			this.records = new DnsRecordList();
		}

		public DnsZoneManager(DnsRecordList records)
		{
			this.records = records;
		}

		public static string GetZoneFileName(string zoneName)
		{
			if (StringUtils.IsEmpty(zoneName))
			{
				throw new DnsServerError(DnsError.DnsNameErrorCode);
			}

			string result = String.Empty;
			foreach (char c in zoneName)
			{
				switch (c)
				{
					case '.': result += "_"; break;
					case '_': result += "__"; break;
					default: result += c; break;
				}
			}
			return result + DnzZoneFileExt;
		}

		public static string GetZoneName(string fileName)
		{
			if (StringUtils.IsEmpty(fileName))
			{
				throw new DnsServerError(DnsError.DnsNameErrorCode);
			}

			string result = String.Empty;
			int cnt = 0;

			foreach (char c in fileName)
			{
				if (c == '.') break;

				if (c == '_')
				{
					cnt++;
					if (cnt == 2)
					{
						cnt = 0;
						result += '_';
					}
				}
				else if (cnt == 1)
				{
					cnt = 0;
					result += '.';
					result += c;
				}
				else
				{
					cnt = 0;
					result += c;
				}
			}

			return result;
		}

		public void Load(string zoneName)
		{
			string fileName = FileUtils.AddTrailingBackSlash(ZonePath) + GetZoneFileName(zoneName);
			if (File.Exists(fileName))
			{
				LoadFromFile(fileName);
			}
		}

		public void LoadFromFile(string fileName)
		{
			Records.Clear();

			using (Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read))
			{
				StringCollectionEx data = new StringCollectionEx();

				data.LoadFromStream(stream, String.Empty);

				DnsRecordPersisterFactory factory = new DnsRecordPersisterFactory();
				foreach (string s in data)
				{
					DnsRecord record = factory.LoadRecord(s);
					Records.Add(record);
				}
			}
		}

		public void Save(string zoneName)
		{
			if (Records.Count == 0) return;

			Directory.CreateDirectory(FileUtils.AddTrailingBackSlash(ZonePath));

			string fileName = FileUtils.AddTrailingBackSlash(ZonePath) + GetZoneFileName(zoneName);
			SaveToFile(fileName);
		}

		public void SaveToFile(string fileName)
		{
			if (Records.Count == 0) return;

			using (Stream stream = new FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.Read))
			{
				StringCollectionEx data = new StringCollectionEx();

				DnsRecordPersisterFactory factory = new DnsRecordPersisterFactory();

				foreach (DnsRecord rec in Records)
				{
					data.Add(factory.SaveRecord(rec));
				}

				data.SaveToStream(stream, String.Empty);
			}
		}

		public string[] ListZones()
		{
			string[] list = Directory.GetFileSystemEntries(FileUtils.AddTrailingBackSlash(ZonePath), "*" + DnzZoneFileExt);

			StringCollectionEx result = new StringCollectionEx();

			foreach (string s in list)
			{
				FileInfo info = new FileInfo(s);
				if ((info.Attributes & FileAttributes.Directory) == 0)
				{
					result.Add(GetZoneName(info.Name));
				}
			}

			return result.ToArray();
		}

		public string ZonePath
		{
			get { return zonePath; }
			set { zonePath = value; }
		}

		public DnsRecordList Records
		{
			get { return records; }
		}
	}

	[ToolboxBitmap(typeof(DnsFileHandler), "Resources.DnsFileHandler.bmp")]
	public class DnsFileHandler : Component
	{
		DnsServer server;
		DnsZoneManager zoneManager;
		string handedZonesPath;
		string cachedZonesPath;

		string EncodeCacheFileName(string name, int recordType)
		{
			return DnsUtils.GetRecordTypeStr(recordType) + "_" + DnsZoneManager.GetZoneFileName(name);
		}

		void DoGetHandedRecords(object sender, DnsRecordsEventArgs e)
		{
			DnsZoneManager zoneMgr = new DnsZoneManager(e.Records);
			zoneMgr.ZonePath = HandedZonesPath;

			lock (this)
			{
				zoneMgr.Load(e.Name);
			}
		}

		void DoGetCachedRecords(object sender, DnsCacheEventArgs e)
		{
			lock (this)
			{
				string fileName = FileUtils.AddTrailingBackSlash(CachedZonesPath) + EncodeCacheFileName(e.Name, e.RecordType);

				if (File.Exists(fileName))
				{
					FileInfo fileInfo = new FileInfo(fileName);

					using (Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read))
					{
						StringCollectionEx data = new StringCollectionEx();

						data.LoadFromStream(stream, String.Empty);

						DnsCachePersister persister = new DnsCachePersister();
						e.CacheEntry = persister.Load(data);

						if (e.CacheEntry != null)
						{
							e.CacheEntry.CreatedOn = fileInfo.LastWriteTime;
						}
					}
				}
			}
		}

		void DoAddCachedRecords(object sender, DnsCacheEventArgs e)
		{
			lock (this)
			{
				string fileName = FileUtils.AddTrailingBackSlash(CachedZonesPath) + EncodeCacheFileName(e.Name, e.RecordType);

				Directory.CreateDirectory(CachedZonesPath);

				using (Stream stream = new FileStream(fileName, FileMode.Create, FileAccess.Write, FileShare.Read))
				{
					StringCollectionEx data = new StringCollectionEx();

					DnsCachePersister persister = new DnsCachePersister();
					persister.Save(e.CacheEntry, data);

					data.SaveToStream(stream, String.Empty);
				}
			}
		}

		void DoDeleteCachedRecords(object sender, DnsDeleteCacheEventArgs e)
		{
			lock (this)
			{
				string fileName = FileUtils.AddTrailingBackSlash(CachedZonesPath) + EncodeCacheFileName(e.Name, e.RecordType);
				if (File.Exists(fileName))
				{
					File.Delete(fileName);
				}
			}
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				CleanEventHandlers();
			}
			base.Dispose(disposing);
		}

		protected virtual void CleanEventHandlers()
		{
			if (server != null)
			{
				server.GetHandedRecords -= new DnsRecordsEventHandler(DoGetHandedRecords);
				server.GetCachedRecords -= new DnsCacheEventHandler(DoGetCachedRecords);
				server.AddCachedRecords -= new DnsCacheEventHandler(DoAddCachedRecords);
				server.DeleteCachedRecords -= new DnsDeleteCacheEventHandler(DoDeleteCachedRecords);
			}
		}

		protected virtual void InitEventHandlers()
		{
			if (server != null)
			{
				server.GetHandedRecords += new DnsRecordsEventHandler(DoGetHandedRecords);
				server.GetCachedRecords += new DnsCacheEventHandler(DoGetCachedRecords);
				server.AddCachedRecords += new DnsCacheEventHandler(DoAddCachedRecords);
				server.DeleteCachedRecords += new DnsDeleteCacheEventHandler(DoDeleteCachedRecords);
			}
		}


		public DnsFileHandler() : base()
		{
			this.zoneManager = new DnsZoneManager();
			this.handedZonesPath = "Handed";
			this.cachedZonesPath = "Cached";
		}

		public DnsZoneManager ZoneManager
		{
			get { return zoneManager; }
		}

		[DefaultValue(null)]
		public DnsServer Server
		{
			get { return server; }
			set
			{
				CleanEventHandlers();
				server = value;
				InitEventHandlers();
			}
		}

		[DefaultValue("Handed")]
		public string HandedZonesPath
		{
			get { return handedZonesPath; }
			set
			{
				lock (this)
				{
					handedZonesPath = value;
				}
			}
		}

		[DefaultValue("Cached")]
		public string CachedZonesPath
		{
			get { return cachedZonesPath; }
			set
			{
				lock (this)
				{
					cachedZonesPath = value;
				}
			}
		}
	}
}
