#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2013 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;

namespace CleverComponents.InetSuite
{

    [ToolboxBitmap(typeof(FtpFileHandler), "Resources.FtpFileHandler.bmp")]
    public class FtpFileHandler : Component
    {
        FtpServer server;

        FtpFileInfo GetFtpFileInfo(FileInfo info)
        {
            FtpFileInfo ftpInfo = new FtpFileInfo();

            ftpInfo.FileName = info.Name;


            if ((info.Attributes & FileAttributes.Directory) > 0)
                ftpInfo.IsDirectory = true;
            else
                ftpInfo.Size = info.Length;

            if ((info.Attributes & FileAttributes.ReadOnly) > 0)
                ftpInfo.IsReadOnly = true;

            ftpInfo.ModifiedDate = info.LastWriteTime;

            return ftpInfo;
        }

        void DoCreateDir(object sender, FtpNameEventArgs e)
        {
            if (Directory.Exists(e.Name))
            {
                e.Success = false;
                e.ErrorMessage = "The directory " + e.Name + " already exists.";
                return;
            }
            try
            {
                Directory.CreateDirectory(e.Name);
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }

        void DoDelete(object sender, FtpNameEventArgs e)
        {
            try
            {
                FileInfo info = new FileInfo(e.Name);
                if ((info.Attributes & FileAttributes.Directory) > 0)
                {
                    Directory.Delete(e.Name, false);
                }
                else
                {
                    if (File.Exists(e.Name))
					{
                        File.Delete(e.Name);
                    }
                    else
					{
                        e.Success = false;
                        e.ErrorMessage = "The file doesn't exist.";
                    }
                }
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }

        void DoRename(object sender, FtpRenameEventArgs e)
        {
            try
            {
                if (File.Exists(e.CurrentName))
                {
                    File.Move(e.CurrentName, e.NewName);
                }
                else
                {
                    Directory.Move(e.CurrentName, e.NewName);
                }
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }

        void DoPutFile(object sender, FtpPutFileEventArgs e)
        {
            try
            {
                e.Stream = new FileStream(e.FileName, e.Overwrite ? FileMode.Create : FileMode.Open,
                    FileAccess.Write, FileShare.Read);
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }

        void DoGetFile(object sender, FtpGetFileEventArgs e)
        {
            try
            {
                e.Stream = new FileStream(e.FileName, FileMode.Open, FileAccess.Read, FileShare.Read);
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }

        void DoGetFileList(object sender, FtpFileListEventArgs e)
        {
            try
            {
                if (Directory.Exists(e.PathName))
                {
                    DirectoryInfo d = new DirectoryInfo(FileUtils.AddTrailingBackSlash(e.PathName));
                    if ((d.Parent == null) && StringUtils.IsEmpty(e.FileMask))
                    {
                        FtpFileInfo ftpInfo = new FtpFileInfo();
                        e.FileList.Add(ftpInfo);

                        ftpInfo.FileName = e.PathName;
                        ftpInfo.IsDirectory = true;
                    }
                    else
                    {
                        string[] list = Directory.GetFileSystemEntries(e.PathName, e.FileMask);
                        foreach (string s in list)
                        {
                            FileInfo info = new FileInfo(s);

                            if (e.IncludeHidden || ((info.Attributes & FileAttributes.Hidden) == 0))
                            {
                                e.FileList.Add(GetFtpFileInfo(info));
                            }
                        }
                    }
                }
                else if (File.Exists(e.PathName))
                {
                    e.FileList.Add(GetFtpFileInfo(new FileInfo(e.PathName)));
                }
            }
            catch (Exception ex)
            {
                e.Success = false;
                e.ErrorMessage = ex.Message;
            }
        }


        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                CleanEventHandlers();
            }
            base.Dispose(disposing);
        }

        protected virtual void CleanEventHandlers()
        {
            if (server != null)
            {
                server.CreateDir -= new FtpNameEventHandler(DoCreateDir);
                server.Delete -= new FtpNameEventHandler(DoDelete);
                server.Rename -= new FtpRenameEventHandler(DoRename);
                server.PutFile -= new FtpPutFileEventHandler(DoPutFile);
                server.GetFile -= new FtpGetFileEventHandler(DoGetFile);
                server.GetFileList -= new FtpFileListEventHandler(DoGetFileList);
            }
        }

        protected virtual void InitEventHandlers()
        {
            if (server != null)
            {
                server.CreateDir += new FtpNameEventHandler(DoCreateDir);
                server.Delete += new FtpNameEventHandler(DoDelete);
                server.Rename += new FtpRenameEventHandler(DoRename);
                server.PutFile += new FtpPutFileEventHandler(DoPutFile);
                server.GetFile += new FtpGetFileEventHandler(DoGetFile);
                server.GetFileList += new FtpFileListEventHandler(DoGetFileList);
            }
        }


        public FtpFileHandler() : base() { }

        [DefaultValue(null)]
        public FtpServer Server
        {
            get { return server; }
            set
            {
                CleanEventHandlers();
                server = value;
                InitEventHandlers();
            }
        }
    }
}
