#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2013 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;

namespace CleverComponents.InetSuite
{

	public class Imap4MailBoxStore
	{
		public const string LoadError = "Cannot load mailbox settings";
		public static int Version = 1;

		public Imap4MailBoxStore(Imap4MailBoxItem mailBox)
		{
			MailBox = mailBox;
		}

		public virtual void Load(Stream stream)
		{
			NonClosingBinaryReader reader = new NonClosingBinaryReader(stream);
			if (reader.ReadInt32() != Version)
			{
				throw new StreamError(LoadError);
			}
			MailBox.IsSubscribed = reader.ReadBool();
			MailBox.UIDNext = reader.ReadInt32();
			MailBox.UIDValidity = reader.ReadString();
			int counter = reader.ReadInt32();
			if (counter > 0)
			{
				MailBox.Data = (Object)counter;
			}
			else
			{
				MailBox.Data = null;
			}
		}

		public virtual void Save(Stream stream)
		{
			NonClosingBinaryWriter writer = new NonClosingBinaryWriter(stream);
			writer.Write(Version);
			writer.Write(MailBox.IsSubscribed);
			writer.Write(MailBox.UIDNext);
			writer.Write(MailBox.UIDValidity);
			if (MailBox.Data != null)
			{
				writer.Write((int)MailBox.Data);
			}
			else
			{
				writer.Write((int)0);
			}
		}

		public Imap4MailBoxItem MailBox { get; }
	}

	public class Imap4MessageStore
	{
		public const string LoadError = "Cannot load messages";
		public static int Version = 2;

		Imap4MessageList messages;

		public Imap4MessageStore(Imap4MessageList messages)
		{
			this.messages = messages;
		}

		public virtual void Load(Stream stream)
		{
			NonClosingBinaryReader reader = new NonClosingBinaryReader(stream);
			if (reader.ReadInt32() != Version)
			{
				throw new StreamError(LoadError);
			}
			Messages.Clear();
			int cnt = reader.ReadInt32();
			for (int i = 0; i < cnt; i++)
			{
				Imap4MessageItem item = new Imap4MessageItem();
				Messages.Add(item);
				item.Name = reader.ReadString();
				item.UID = reader.ReadInt32();
				item.Flags = (MailMessageFlags)reader.ReadInt32();
				item.Date = new DateTime(reader.ReadInt64());
			}
		}

		public virtual void Save(Stream stream)
		{
			NonClosingBinaryWriter writer = new NonClosingBinaryWriter(stream);
			writer.Write(Version);

			writer.Write((int)Messages.Count);
			foreach (Imap4MessageItem item in Messages)
			{
				writer.Write(item.Name);
				writer.Write(item.UID);
				writer.Write((int)item.Flags);
				writer.Write(item.Date.Ticks);
			}
		}

		public Imap4MessageList Messages
		{
			get { return messages; }
		}
	}

	[ToolboxBitmap(typeof(Imap4FileHandler), "Resources.Imap4FileHandler.bmp")]
	public class Imap4FileHandler : Component
	{
		public const string ImapMailBoxInfoFile = "imap.dat";
		public const string ImapMessagesInfoFile = "messages.dat";
		public static int MaxTryCount = 1000;

		Imap4Server server;
		string mailBoxDir;
		string mailBoxInfoFile;
		string messagesInfoFile;
		string sharedFolders;


		string GetMailBoxRoot(string userName)
		{
			return FileUtils.AddTrailingBackSlash(MailBoxDir) + FileUtils.AddTrailingBackSlash(userName);
		}

		string MailBoxToPath(string mailBox)
		{
			if (StringUtils.IsEmpty(mailBox) || ("INBOX" == StringUtils.UpperCase(mailBox)))
			{
				return String.Empty;
			}
			return mailBox.Replace(Server.MailBoxSeparator, '\\');
		}

		string GetMailBoxPath(string userName, string mailBox)
		{
			string result = GetMailBoxRoot(userName) + FileUtils.AddTrailingBackSlash(MailBoxToPath(mailBox));
			if (HasSharedFolders() && (!Directory.Exists(result)))
			{
				result = FileUtils.AddTrailingBackSlash(SharedFolders)
					+ FileUtils.AddTrailingBackSlash(MailBoxToPath(mailBox));
			}
			return result;
		}

		bool HasSharedFolders()
		{
			return !StringUtils.IsEmpty(SharedFolders) && Directory.Exists(SharedFolders);
		}

		bool IsSharedFolder(string mailboxPath)
		{
			if (StringUtils.IsEmpty(mailboxPath) || (!HasSharedFolders()))
				return false;

			string sharedPath = StringUtils.UpperCase(SharedFolders);
			return (StringUtils.UpperCase(mailboxPath).IndexOf(sharedPath) == 0);
		}

		void CollectMailBoxes(Imap4MailBoxList mailBoxes, string path, string rootMailBox)
		{
			string[] list = Directory.GetDirectories(path);
			foreach (string s in list)
			{
				DirectoryInfo info = new DirectoryInfo(s);
				Imap4MailBoxItem item = new Imap4MailBoxItem(rootMailBox + info.Name);
				CollectMailBoxInfo(item, s, false);
				mailBoxes.Add(item);

				CollectMailBoxes(mailBoxes, s, info.Name + Server.MailBoxSeparator);
			}
		}

		string GenUidValidity(string mailboxPath)
		{
			if (Directory.Exists(mailboxPath))
			{
				DirectoryInfo info = new DirectoryInfo(mailboxPath);
				return info.LastWriteTime.Ticks.ToString();
			}
			return "0";
		}

		void SetCounter(Imap4MailBoxItem item, int counter)
		{
			item.Data = (Object)counter;
		}

		int GetCounter(Imap4MailBoxItem item)
		{
			if (item.Data != null)
			{
				return (int)item.Data;
			}
			return 0;
		}

		void CollectMailBoxInfo(Imap4MailBoxItem item, string mailboxPath, bool isSubscribed)
		{
			Imap4MailBoxStore store = new Imap4MailBoxStore(item);
			string s = FileUtils.AddTrailingBackSlash(mailboxPath) + MailBoxInfoFile;

			if (!File.Exists(s))
			{
				item.IsSubscribed = isSubscribed;
				item.UIDNext = 1;
				item.UIDValidity = GenUidValidity(mailboxPath);
				SetCounter(item, 0);

				using (Stream stream = new FileStream(s, FileMode.Create, FileAccess.Write))
				{
					store.Save(stream);
				}
			}
			else
			{
				using (Stream stream = new FileStream(s, FileMode.Open, FileAccess.Read))
				{
					store.Load(stream);
				}
			}
		}

		void UpdateMailBoxInfo(string mailBoxPath, Imap4MailBoxItem mailBox)
		{
			Imap4MailBoxStore store = new Imap4MailBoxStore(mailBox);
			using (Stream stream = new FileStream(mailBoxPath, FileMode.Create, FileAccess.Write))
			{
				store.Save(stream);
			}
		}

		void UpdateMessageInfo(string msgInfoFile, string messageFile, MailMessageFlags flags, DateTime date)
		{
			Imap4MessageList storedMessages = new Imap4MessageList();

			if (File.Exists(msgInfoFile))
			{
				using (Stream stream = new FileStream(msgInfoFile, FileMode.Open, FileAccess.Read))
				{
					Imap4MessageStore store = new Imap4MessageStore(storedMessages);
					store.Load(stream);
				}
			}

			Imap4MessageItem storedItem = storedMessages.FindByName(messageFile);
			if (storedItem == null)
			{
				storedItem = new Imap4MessageItem(messageFile);
				storedMessages.Add(storedItem);
			}

			storedItem.Flags = flags;

			try
			{
				FileInfo fileInfo = new FileInfo(messageFile);
				fileInfo.LastWriteTime = date;
			}
			catch (UnauthorizedAccessException) { }
			catch (IOException) { }
			using (Stream stream = new FileStream(msgInfoFile, FileMode.Create, FileAccess.Write))
			{
				Imap4MessageStore store = new Imap4MessageStore(storedMessages);
				store.Save(stream);
			}
		}

		void DoGetMailBoxes(object sender, Imap4GetMailBoxesEventArgs e)
		{
			lock (this)
			{
				string path = GetMailBoxRoot(e.Connection.UserName);

				if (StringUtils.IsEmpty(e.SelectedMailBox))
				{
					if (Directory.Exists(path))
					{
						Imap4MailBoxItem item = new Imap4MailBoxItem("INBOX");
						CollectMailBoxInfo(item, path, true);
						e.MailBoxes.Add(item);

						CollectMailBoxes(e.MailBoxes, path, String.Empty);
					}

					if (HasSharedFolders())
					{
						CollectMailBoxes(e.MailBoxes, FileUtils.AddTrailingBackSlash(SharedFolders), String.Empty);
					}
				}
				else
				{
					path = GetMailBoxPath(e.Connection.UserName, e.SelectedMailBox);
					if (Directory.Exists(path))
					{
						Imap4MailBoxItem item = new Imap4MailBoxItem(e.SelectedMailBox);
						CollectMailBoxInfo(item, path, false);
						e.MailBoxes.Add(item);
					}
				}
			}
		}

		void DoUpdateMailBox(object sender, Imap4UpdateMailBoxEventArgs e)
		{
			lock (this)
			{
				string s = GetMailBoxPath(e.Connection.UserName, e.MailBox.Name) + MailBoxInfoFile;
				if (File.Exists(s))
				{
					UpdateMailBoxInfo(s, e.MailBox);
				}
			}
		}

		void DoCreateMailBox(object sender, Imap4MailBoxEventArgs e)
		{
			lock (this)
			{
				e.Success = false;

				if ("INBOX" == e.MailBox.ToUpper(CultureInfo.InvariantCulture))
					return;

				string path = GetMailBoxPath(e.Connection.UserName, e.MailBox);
				if (Directory.Exists(path) || IsSharedFolder(path))
					return;

				try
				{
					Directory.CreateDirectory(path);
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoDeleteMailBox(object sender, Imap4MailBoxEventArgs e)
		{
			lock (this)
			{
				e.Success = false;

				if ("INBOX" == e.MailBox.ToUpper(CultureInfo.InvariantCulture))
					return;

				string path = GetMailBoxPath(e.Connection.UserName, e.MailBox);
				if ((!Directory.Exists(path)) || IsSharedFolder(path))
					return;

				string[] dirs = Directory.GetDirectories(path);
				if ((dirs != null) && (dirs.Length > 0))
					return;

				try
				{
					Directory.Delete(path, true);
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoRenameMailBox(object sender, Imap4RenameMailBoxEventArgs e)
		{
			lock (this)
			{
				e.Success = false;

				if (("INBOX" == e.CurrentName.ToUpper(CultureInfo.InvariantCulture))
					|| ("INBOX" == e.NewName.ToUpper(CultureInfo.InvariantCulture)))
					return;

				string currentPath = GetMailBoxPath(e.Connection.UserName, e.CurrentName);
				string newPath = GetMailBoxPath(e.Connection.UserName, e.NewName);

				if ((!Directory.Exists(currentPath)) || IsSharedFolder(currentPath)
					|| (Directory.Exists(newPath)) || IsSharedFolder(newPath))
					return;

				try
				{
					Directory.Move(currentPath, newPath);
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoGetMessages(object sender, Imap4MessagesEventArgs e)
		{
			lock (this)
			{
				e.Success = false;
				try
				{
					string mailBoxPath = GetMailBoxPath(e.Connection.UserName, e.MailBox);

					if (!Directory.Exists(mailBoxPath))
						return;

					string msgInfoFile = mailBoxPath + MessagesInfoFile;
					Imap4MessageList storedMessages = new Imap4MessageList();

					if (File.Exists(msgInfoFile))
					{
						using (Stream stream = new FileStream(msgInfoFile, FileMode.Open, FileAccess.Read))
						{
							Imap4MessageStore store = new Imap4MessageStore(storedMessages);
							store.Load(stream);
						}
					}

					string[] list = Directory.GetFiles(mailBoxPath, "*.MSG");

					for (int i = storedMessages.Count - 1; i > -1; i--)
					{
						Imap4MessageItem item = storedMessages[i];
						if (Array.IndexOf(list, item.Name) < 0)
						{
							storedMessages.RemoveAt(i);
						}
					}

					foreach (string s in list)
					{
						Imap4MessageItem storedItem = storedMessages.FindByName(s);
						Imap4MessageItem item = new Imap4MessageItem(s);

						if (storedItem != null)
						{
							item.UID = storedItem.UID;
							item.Flags = storedItem.Flags;
						}

						FileInfo fileInfo = new FileInfo(s);
						item.Date = fileInfo.LastWriteTime;

						e.Messages.Add(item);
					}
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoUpdateMessages(object sender, Imap4MessagesEventArgs e)
		{
			lock (this)
			{
				e.Success = false;
				try
				{
					string mailBoxPath = GetMailBoxPath(e.Connection.UserName, e.MailBox);
					string msgInfoFile = mailBoxPath + MessagesInfoFile;

					Imap4MessageList storedMessages = new Imap4MessageList();

					if (File.Exists(msgInfoFile))
					{
						using (Stream stream = new FileStream(msgInfoFile, FileMode.Open, FileAccess.Read))
						{
							Imap4MessageStore store = new Imap4MessageStore(storedMessages);
							store.Load(stream);
						}
					}

					foreach (Imap4MessageItem msg in e.Messages)
					{
						if (!File.Exists(msg.Name))
							continue;

						Imap4MessageItem storedItem = storedMessages.FindByName(msg.Name);
						if (storedItem == null)
						{
							storedItem = new Imap4MessageItem(msg.Name);

							storedItem.UID = msg.UID;
							storedMessages.Add(storedItem);
						}
						else if (storedItem.UID < 0)
						{
							storedItem.UID = msg.UID;
						}

						storedItem.Flags = msg.Flags;

						try
						{
							FileInfo fileInfo = new FileInfo(msg.Name);
							fileInfo.LastWriteTime = msg.Date;
						}
						catch (UnauthorizedAccessException) { }
						catch (IOException) { }
					}

					using (Stream stream = new FileStream(msgInfoFile, FileMode.Create, FileAccess.Write))
					{
						Imap4MessageStore store = new Imap4MessageStore(storedMessages);
						store.Save(stream);
					}
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoDeleteMessage(object sender, Imap4MessageEventArgs e)
		{
			lock (this)
			{
				e.Success = false;
				try
				{
					if (File.Exists(e.Message.Name))
					{
						File.Delete(e.Message.Name);
						e.Success = true;
					}
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoGetMessageSource(object sender, Imap4GetMessageSourceEventArgs e)
		{
			lock (this)
			{
				e.Success = false;
				try
				{
					if (File.Exists(e.MessageName))
					{
						using (Stream stream = new FileStream(e.MessageName, FileMode.Open, FileAccess.Read))
						{
							e.MessageSource.LoadFromStream(stream, String.Empty);
						}
						e.Success = true;
					}
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}

		void DoAppendMessage(object sender, Imap4AppendMessageEventArgs e)
		{
			lock (this)
			{
				e.Success = false;
				try
				{
					string mailBoxPath = GetMailBoxPath(e.Connection.UserName, e.MailBox);
					Imap4MailBoxItem mailBoxItem = new Imap4MailBoxItem(e.MailBox);
					CollectMailBoxInfo(mailBoxItem, mailBoxPath, false);
					int counter = GetCounter(mailBoxItem);
					string msgFileName = String.Empty;

					int i = 0;
					while (true)
					{
						counter++;
						msgFileName = mailBoxPath + String.Format("MAIL{0,8}.MSG", counter).Replace(' ', '0');

						try
						{
							if (!File.Exists(msgFileName))
							{
								using (Stream stream = new FileStream(msgFileName, FileMode.CreateNew, FileAccess.Write))
								{
									e.MessageSource.SaveToStream(stream, String.Empty);
									break;
								}
							}
						}
						catch (IOException) { }
						i++;
						if (i > MaxTryCount)
						{
							throw new Exception("Cannot save new message");
						}
					}

					SetCounter(mailBoxItem, counter);
					UpdateMailBoxInfo(mailBoxPath + MailBoxInfoFile, mailBoxItem);
					UpdateMessageInfo(mailBoxPath + MessagesInfoFile, msgFileName, e.Flags, e.Date);
					e.Success = true;
				}
				catch (UnauthorizedAccessException) { }
				catch (IOException) { }
			}
		}


		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				CleanEventHandlers();
			}
			base.Dispose(disposing);
		}

		protected virtual void CleanEventHandlers()
		{
			if (server != null)
			{
				server.GetMailBoxes -= new Imap4GetMailBoxesEventHandler(DoGetMailBoxes);
				server.UpdateMailBox -= new Imap4UpdateMailBoxEventHandler(DoUpdateMailBox);
				server.CreateMailBox -= new Imap4MailBoxEventHandler(DoCreateMailBox);
				server.DeleteMailBox -= new Imap4MailBoxEventHandler(DoDeleteMailBox);
				server.RenameMailBox -= new Imap4RenameMailBoxEventHandler(DoRenameMailBox);
				server.GetMessages -= new Imap4MessagesEventHandler(DoGetMessages);
				server.UpdateMessages -= new Imap4MessagesEventHandler(DoUpdateMessages);
				server.DeleteMessage -= new Imap4MessageEventHandler(DoDeleteMessage);
				server.GetMessageSource -= new Imap4GetMessageSourceEventHandler(DoGetMessageSource);
				server.AppendMessage -= new Imap4AppendMessageEventHandler(DoAppendMessage);
			}
		}

		protected virtual void InitEventHandlers()
		{
			if (server != null)
			{
				server.GetMailBoxes += new Imap4GetMailBoxesEventHandler(DoGetMailBoxes);
				server.UpdateMailBox += new Imap4UpdateMailBoxEventHandler(DoUpdateMailBox);
				server.CreateMailBox += new Imap4MailBoxEventHandler(DoCreateMailBox);
				server.DeleteMailBox += new Imap4MailBoxEventHandler(DoDeleteMailBox);
				server.RenameMailBox += new Imap4RenameMailBoxEventHandler(DoRenameMailBox);
				server.GetMessages += new Imap4MessagesEventHandler(DoGetMessages);
				server.UpdateMessages += new Imap4MessagesEventHandler(DoUpdateMessages);
				server.DeleteMessage += new Imap4MessageEventHandler(DoDeleteMessage);
				server.GetMessageSource += new Imap4GetMessageSourceEventHandler(DoGetMessageSource);
				server.AppendMessage += new Imap4AppendMessageEventHandler(DoAppendMessage);
			}
		}


		public Imap4FileHandler() : base()
		{
			this.mailBoxDir = String.Empty;
			this.mailBoxInfoFile = ImapMailBoxInfoFile;
			this.messagesInfoFile = ImapMessagesInfoFile;
			this.sharedFolders = String.Empty;
		}

		[DefaultValue(null)]
		public Imap4Server Server
		{
			get { return server; }
			set
			{
				lock (this)
				{
					CleanEventHandlers();
					server = value;
					InitEventHandlers();
				}
			}
		}

		[DefaultValue("")]
		public string MailBoxDir
		{
			get { return mailBoxDir; }
			set
			{
				lock (this)
				{
					mailBoxDir = value;
				}
			}
		}

		[DefaultValue(ImapMailBoxInfoFile)]
		public string MailBoxInfoFile
		{
			get { return mailBoxInfoFile; }
			set
			{
				lock (this)
				{
					mailBoxInfoFile = value;
				}
			}
		}

		[DefaultValue(ImapMessagesInfoFile)]
		public string MessagesInfoFile
		{
			get { return messagesInfoFile; }
			set
			{
				lock (this)
				{
					messagesInfoFile = value;
				}
			}
		}

		[DefaultValue("")]
		public string SharedFolders
		{
			get { return sharedFolders; }
			set
			{
				lock (this)
				{
					sharedFolders = value;
				}
			}
		}
	}
}
