#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2013 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;

namespace CleverComponents.InetSuite
{

	[ToolboxBitmap(typeof(NntpFileHandler), "Resources.NntpFileHandler.bmp")]
	public class NntpFileHandler : Component
	{
		public static int MaxTryCount = 1000;

		NntpServer server;
		string rootDir;
		int counter;

		string GetGroupPath(string groupName)
		{
			string result = FileUtils.AddTrailingBackSlash(RootDir);
			if (!StringUtils.IsEmpty(groupName))
			{
				result += groupName.Replace('.', '_');
			}

			lock (this)
			{
				if (!Directory.Exists(result))
				{
					Directory.CreateDirectory(result);
				}
			}

			return result;
		}

		NntpArticleItem CreateArticleItem(string fileName)
		{
			string[] words = StringUtils.ExtractWords(Path.GetFileNameWithoutExtension(fileName), new char[] { '_' });
			if (words.Length != 2)
			{
				return null;
			}

			int articleNo = StringUtils.StrToIntDef(words[0], 0);
			string messageID = Encoder.Decode(words[1], EncodeMethod.Base64);

			if ((articleNo < 1) || StringUtils.IsEmpty(messageID))
			{
				return null;
			}

			return new NntpArticleItem(articleNo, messageID);
		}

		void CollectArticles(NntpArticleList articles, string path, DateTime date, bool gmt)
		{
			string[] list = Directory.GetFileSystemEntries(path, "*.MSG");

			foreach (string s in list)
			{
				FileInfo info = new FileInfo(s);

				DateTime fileDate = gmt ? info.LastWriteTimeUtc : info.LastWriteTime;

				if (fileDate >= date)
				{
					NntpArticleItem item = CreateArticleItem(s);
					if (item != null)
					{
						articles.Add(item);
					}
				}
			}
		}

		string GenerateArticleName(int articleNo, string messageID)
		{
			return articleNo.ToString("0000000") + "_" + Encoder.EncodeToString(messageID, EncodeMethod.Base64) + ".MSG";
		}

		void DoGetGroupInfo(object sender, NntpGetGroupInfoEventArgs e)
		{
			string path = GetGroupPath(e.Group.Name);

			NntpArticleList articles = new NntpArticleList();
			CollectArticles(articles, path, DateTime.MinValue, false);

			e.ArticleCount = articles.Count;
			e.FirstArticle = 0;
			e.LastArticle = 0;

			if (articles.Count > 0)
			{
				e.FirstArticle = articles[0].ArticleNo;
				e.LastArticle = articles[articles.Count - 1].ArticleNo;
			}
		}

		void DoGetArticles(object sender, NntpGetArticlesEventArgs e)
		{
			string path = GetGroupPath(e.Group.Name);

			CollectArticles(e.Articles, path, e.Date, e.GMT);
		}

		void DoGetArticleSource(object sender, NntpGetArticleSourceEventArgs e)
		{
			string path = GetGroupPath(e.Group.Name);

			NntpArticleList articles = new NntpArticleList();
			CollectArticles(articles, path, DateTime.MinValue, false);

			NntpArticleItem article = null;
			if (e.ArticleNo > 0)
			{
				article = articles.FindArticle(e.ArticleNo);
			}
			else if (!StringUtils.IsEmpty(e.MessageID))
			{
				article = articles.FindArticle(e.MessageID);
			}

			if (article == null)
			{
				e.Success = false;
				return;
			}

			e.ArticleNo = article.ArticleNo;
			e.MessageID = article.MessageID;
			string fileName = FileUtils.AddTrailingBackSlash(path) + GenerateArticleName(e.ArticleNo, e.MessageID);

			using (FileStream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read))
			{
				e.ArticleSource.LoadFromStream(stream, String.Empty);
			}
			e.Success = true;
		}

		void DoArticleReceived(object sender, NntpArticleReceivedEventArgs e)
		{
			string path = FileUtils.AddTrailingBackSlash(GetGroupPath(e.Group.Name));

			lock (this)
			{
				int i = 0;
				while (true)
				{
					counter++;
					string msgFileName = path + GenerateArticleName(counter, e.MessageID);

					try
					{
						if (!File.Exists(msgFileName))
						{
							using (Stream stream = new FileStream(msgFileName, FileMode.CreateNew, FileAccess.Write))
							{
								e.ArticleSource.SaveToStream(stream, String.Empty);
								e.Success = true;
								break;
							}
						}
					}
					catch (IOException) { }
					i++;
					if (i > MaxTryCount)
					{
						throw new Exception("Cannot save new message");
					}
				}
			}
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				CleanEventHandlers();
			}
			base.Dispose(disposing);
		}

		protected virtual void CleanEventHandlers()
		{
			if (server != null)
			{
				server.GetGroupInfo -= new NntpGetGroupInfoEventHandler(DoGetGroupInfo);
				server.GetArticles -= new NntpGetArticlesEventHandler(DoGetArticles);
				server.GetArticleSource -= new NntpGetArticleSourceEventHandler(DoGetArticleSource);
				server.ArticleReceived -= new NntpArticleReceivedEventHandler(DoArticleReceived);
			}
		}

		protected virtual void InitEventHandlers()
		{
			if (server != null)
			{
				server.GetGroupInfo += new NntpGetGroupInfoEventHandler(DoGetGroupInfo);
				server.GetArticles += new NntpGetArticlesEventHandler(DoGetArticles);
				server.GetArticleSource += new NntpGetArticleSourceEventHandler(DoGetArticleSource);
				server.ArticleReceived += new NntpArticleReceivedEventHandler(DoArticleReceived);
			}
		}


		public NntpFileHandler() : base()
		{
			rootDir = "";
			counter = 0;
		}

		[DefaultValue(null)]
		public NntpServer Server
		{
			get { return server; }
			set
			{
				CleanEventHandlers();
				server = value;
				InitEventHandlers();
			}
		}

		[DefaultValue("")]
		public string RootDir
		{
			get { return rootDir; }
			set
			{
				lock (this)
				{
					rootDir = value;
				}
			}
		}

		[DefaultValue(0)]
		public int Counter
		{
			get { return counter; }
			set
			{
				lock (this)
				{
					counter = value;
				}
			}
		}
	}
}
