#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2013 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System.IO;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;

namespace CleverComponents.InetSuite
{

	public delegate void Pop3LoadMessageEventHandler(object sender, Pop3LoadMessageEventArgs e);

	public class Pop3LoadMessageEventArgs : Pop3ConnectionEventArgs
	{
		string mailBoxPath;
		string messageFile;
		bool canLoad;

		public Pop3LoadMessageEventArgs(Pop3CommandConnection connection,
			string mailBoxPath, string messageFile, bool canLoad) : base(connection)
		{
			this.mailBoxPath = mailBoxPath;
			this.messageFile = messageFile;
			this.canLoad = canLoad;
		}

		public string MailBoxPath
		{
			get { return mailBoxPath; }
		}

		public string MessageFile
		{
			get { return messageFile; }
		}

		public bool CanLoad
		{
			get { return canLoad; }
			set { canLoad = value; }
		}
	}

	public class Pop3FileMessageItem : Pop3MessageItem
	{
		string path;

		public Pop3FileMessageItem(string uid, long size, string path) : base(uid, size)
		{
			this.path = path;
		}

		public string Path
		{
			get { return path; }
			set { path = value; }
		}
	}

	[ToolboxBitmap(typeof(Pop3FileHandler), "Resources.Pop3FileHandler.bmp")]
	public class Pop3FileHandler : Component
	{
		Pop3Server server;
		string mailBoxDir;

		static readonly object messageLoading = new object();


		void DoStateChanged(object sender, Pop3ConnectionEventArgs e)
		{
			if (e.Connection.ConnectionState != Pop3ConnectionState.Update)
				return;

			foreach (Pop3FileMessageItem item in e.Connection.MailBox)
			{
				if (item.IsDeleted)
				{
					File.Delete(item.Path);
				}
			}
		}

		void DoGetMailBoxInfo(object sender, Pop3MailBoxEventArgs e)
		{
			string path = GetMailBoxPath(e.Connection.UserName);
			e.MailBox.Clear();

			string[] list = Directory.GetFileSystemEntries(path, "*.MSG");
			foreach (string s in list)
			{
				FileInfo info = new FileInfo(s);
				if ((info.Attributes & FileAttributes.Directory) == 0)
				{
					Pop3LoadMessageEventArgs loadArgs = new Pop3LoadMessageEventArgs(
						e.Connection, path, info.Name, true);

					OnMessageLoading(loadArgs);
					if (loadArgs.CanLoad)
					{
						e.MailBox.Add(new Pop3FileMessageItem(
							GetMessageUid(info.Name, info.Length), info.Length, path + info.Name));
					}
				}
			}
		}

		void DoRetrieve(object sender, Pop3RetrieveEventArgs e)
		{
			try
			{
				string s = ((Pop3FileMessageItem)e.Connection.MailBox[e.MessageNo - 1]).Path;
				using (Stream stream = new FileStream(s, FileMode.Open, FileAccess.Read, FileShare.Read))
				{
					e.Message.LoadFromStream(stream, string.Empty);
				}
				e.Success = true;
			}
			catch
			{
				e.Message.Clear();
				e.Success = false;
			}
		}

		string GetMailBoxPath(string userName)
		{
			return FileUtils.AddTrailingBackSlash(MailBoxDir) + FileUtils.AddTrailingBackSlash(userName);
		}

		string GetMessageUid(string fileName, long fileSize)
		{
			string result = fileName.Replace('.', ':').ToUpper(CultureInfo.InvariantCulture);
			result += ":" + fileSize.ToString();
			return result.Replace(' ', '_');
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				CleanEventHandlers();
			}
			base.Dispose(disposing);
		}

		protected virtual void CleanEventHandlers()
		{
			if (server != null)
			{
				server.StateChanged -= new Pop3ConnectionEventHandler(DoStateChanged);
				server.GetMailBoxInfo -= new Pop3MailBoxEventHandler(DoGetMailBoxInfo);
				server.Retrieve -= new Pop3RetrieveEventHandler(DoRetrieve);
			}
		}

		protected virtual void InitEventHandlers()
		{
			if (server != null)
			{
				server.StateChanged += new Pop3ConnectionEventHandler(DoStateChanged);
				server.GetMailBoxInfo += new Pop3MailBoxEventHandler(DoGetMailBoxInfo);
				server.Retrieve += new Pop3RetrieveEventHandler(DoRetrieve);
			}
		}

		protected virtual void OnMessageLoading(Pop3LoadMessageEventArgs e)
		{
			Pop3LoadMessageEventHandler handler = (Pop3LoadMessageEventHandler)Events[messageLoading];
			if (handler != null)
			{
				handler(this, e);
			}
		}


		public Pop3FileHandler() : base()
		{
			mailBoxDir = "";
		}

		[DefaultValue(null)]
		public Pop3Server Server
		{
			get { return server; }
			set
			{
				CleanEventHandlers();
				server = value;
				InitEventHandlers();
			}
		}

		[DefaultValue("")]
		public string MailBoxDir
		{
			get { return mailBoxDir; }
			set
			{
				lock (this)
				{
					mailBoxDir = value;
				}
			}
		}

		public event Pop3LoadMessageEventHandler MessageLoading
		{
			add
			{
				Events.AddHandler(messageLoading, value);
			}
			remove
			{
				Events.RemoveHandler(messageLoading, value);
			}
		}
	}
}
