﻿#region Copyright (C) Clever Components
//Clever Internet Suite
//Copyright (C) 2021 Clever Components
//All Rights Reserved
//www.CleverComponents.com
#endregion

using System;
using System.IO;
using System.ComponentModel;
using System.Drawing;
using System.Collections.Generic;

namespace CleverComponents.InetSuite
{
	[ToolboxBitmap(typeof(SFtpFileHandler), "Resources.SFtpFileHandler.bmp")]
	public class SFtpFileHandler : Component
	{
		SFtpServer server;
		
		void DoMakeDir(object sender, SFtpPathAttrsEventArgs e)
		{
			if (Directory.Exists(e.Path))
			{
				throw new SFtpServerError(SFtpServerError.SFtpDirectoryExistsError, SFtpStatus.SSH_FX_FAILURE);
			}

			try
			{
				Directory.CreateDirectory(e.Path);
			}
			catch (Exception ex)
			{
				throw new SFtpServerError(ex.Message, SFtpStatus.SSH_FX_FAILURE, ex);
			}
		}

		void DoRemoveDir(object sender, SFtpPathEventArgs e)
		{
			try
			{
				Directory.Delete(e.Path, false);
			}
			catch (Exception ex)
			{
				throw new SFtpServerError(ex.Message, SFtpStatus.SSH_FX_FAILURE, ex);
			}
		}

		void DoGetFileList(object sender, SFtpFileListEventArgs e)
		{
			e.FileList = new SFtpFileSystemEnumerator(e.Path, e.FileMask);
		}

		void DoOpenFile(object sender, SFtpFileEventArgs e)
		{
			e.FilePersister = new SFtpFileSystemPersister(e.Path, e.Flags);
		}

		void DoRemoveFile(object sender, SFtpPathEventArgs e)
		{
			try
			{
				if (!File.Exists(e.Path))
				{
					throw new SFtpServerError(SFtpServerError.SFtpNoSuchFile, SFtpStatus.SSH_FX_FAILURE);
				}

				File.Delete(e.Path);
			}
			catch (Exception ex)
			{
				throw new SFtpServerError(ex.Message, SFtpStatus.SSH_FX_FAILURE, ex);
			}
		}

		void DoRename(object sender, SFtpRenameEventArgs e)
		{
			try
			{
				if (File.Exists(e.OldPath))
				{
					File.Move(e.OldPath, e.NewPath);
				}
				else
				{
					Directory.Move(e.OldPath, e.NewPath);
				}
			}
			catch (Exception ex)
			{
				throw new SFtpServerError(ex.Message, SFtpStatus.SSH_FX_FAILURE, ex);
			}
		}

		void DoMakeLink(object sender, SFtpMakeLinkEventArgs e)
		{
			throw new SFtpServerError(SFtpServerError.SFtpCommandUnsupported, SFtpStatus.SSH_FX_OP_UNSUPPORTED);//TODO
		}

		void DoReadLink(object sender, SFtpReadLinkEventArgs e)
		{
			throw new SFtpServerError(SFtpServerError.SFtpCommandUnsupported, SFtpStatus.SSH_FX_OP_UNSUPPORTED);//TODO
		}

		void DoSetAttributes(object sender, SFtpPathAttrsEventArgs e)
		{
			try
			{
				FileInfo fi = new FileInfo(e.Path);

				if ((e.Attributes.Flags & SFtpFileAttrs.SSH_FILEXFER_ATTR_ACMODTIME) != 0)
				{
					fi.LastAccessTimeUtc = e.Attributes.AccessTime;
					fi.LastWriteTimeUtc = e.Attributes.ModifyTime;
				}

				if ((e.Attributes.Flags & SFtpFileAttrs.SSH_FILEXFER_ATTR_PERMISSIONS) != 0)
				{
					fi.IsReadOnly = !((e.Attributes.Permissions & SFtpFilePermissions.OwnerWrite) != 0
						|| (e.Attributes.Permissions & SFtpFilePermissions.GroupWrite) != 0
						|| (e.Attributes.Permissions & SFtpFilePermissions.OtherWrite) != 0);
				}
			}
			catch (Exception ex)
			{
				throw new SFtpServerError(ex.Message, SFtpStatus.SSH_FX_FAILURE, ex);
			}
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				CleanEventHandlers();
			}
			base.Dispose(disposing);
		}

		protected virtual void CleanEventHandlers()
		{
			if (Server != null)
			{
				Server.MakeDir -= new SFtpPathAttrsEventHandler(DoMakeDir);
				Server.RemoveDir -= new SFtpPathEventHandler(DoRemoveDir);
				Server.GetFileList -= new SFtpFileListEventHandler(DoGetFileList);
				Server.OpenFile -= new SFtpFileEventHandler(DoOpenFile);
				Server.RemoveFile -= new SFtpPathEventHandler(DoRemoveFile);
				Server.Rename -= new SFtpRenameEventHandler(DoRename);
				Server.MakeLink -= new SFtpMakeLinkEventHandler(DoMakeLink);
				Server.ReadLink -= new SFtpReadLinkEventHandler(DoReadLink);
				Server.SetAttributes -= new SFtpPathAttrsEventHandler(DoSetAttributes);
			}
		}

		protected virtual void InitEventHandlers()
		{
			if (Server != null)
			{
				Server.MakeDir += new SFtpPathAttrsEventHandler(DoMakeDir);
				Server.RemoveDir += new SFtpPathEventHandler(DoRemoveDir);
				Server.GetFileList += new SFtpFileListEventHandler(DoGetFileList);
				Server.OpenFile += new SFtpFileEventHandler(DoOpenFile);
				Server.RemoveFile += new SFtpPathEventHandler(DoRemoveFile);
				Server.Rename += new SFtpRenameEventHandler(DoRename);
				Server.MakeLink += new SFtpMakeLinkEventHandler(DoMakeLink);
				Server.ReadLink += new SFtpReadLinkEventHandler(DoReadLink);
				Server.SetAttributes += new SFtpPathAttrsEventHandler(DoSetAttributes);
			}
		}

		public SFtpFileHandler() : base() {	}

		[DefaultValue(null)]
		public SFtpServer Server
		{
			get { return server; }
			set
			{
				CleanEventHandlers();
				server = value;
				InitEventHandlers();
			}
		}
	}

	public class SFtpFileSystemEnumerator : SFtpFileEnumerator
	{
		IEnumerator<string> enumerator;

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				enumerator?.Dispose();
			}
			enumerator = null;

			base.Dispose(disposing);
		}

		public SFtpFileSystemEnumerator(string path, string fileMask) : base(path, fileMask)
		{
			if (StringUtils.IsEmpty(fileMask))
			{
				path = FileUtils.RemoveTrailingPathDelimiter(path);
				fileMask = System.IO.Path.GetFileName(path);
				path = System.IO.Path.GetDirectoryName(path);
			}
			enumerator = Directory.EnumerateFileSystemEntries(path, fileMask, SearchOption.TopDirectoryOnly).GetEnumerator();
		}

		public override SFtpFileInfo Current
		{
			get
			{
				SFtpFileInfo info = new SFtpFileInfo();
				FileInfo fi = new FileInfo(enumerator.Current);
				
				info.FileName =  fi.Name;
				info.Attributes.SetDate(fi.LastWriteTimeUtc, fi.LastWriteTime);

				if (fi.IsReadOnly)
				{
					info.Attributes.SetPermissions(SFtpFilePermissions.OwnerRead | SFtpFilePermissions.GroupRead | SFtpFilePermissions.OtherRead);
				}
				else
				{
					info.Attributes.SetPermissions(
						SFtpFilePermissions.OwnerRead | SFtpFilePermissions.OwnerWrite | SFtpFilePermissions.OwnerExec |
						SFtpFilePermissions.GroupRead | SFtpFilePermissions.GroupWrite | SFtpFilePermissions.GroupExec |
						SFtpFilePermissions.OtherRead | SFtpFilePermissions.OtherWrite | SFtpFilePermissions.OtherExec
					);
				}

				if ((fi.Attributes & FileAttributes.Directory) != 0)
				{
					info.Attributes.SetIsDir(true);
				}
				else
				{
					info.Attributes.SetSize(fi.Length);
				}

				if ((fi.Attributes & FileAttributes.ReparsePoint) != 0)
				{
					info.Attributes.SetIsLink(true);
				}
				
				return info;
			}
		}

		public override bool MoveNext()
		{
			return enumerator.MoveNext();
		}

		public override void Reset()
		{
			enumerator.Reset();
		}
	}

	public class SFtpFileSystemPersister : SFtpFilePersister
	{
		FileStream stream;
		bool appendMode;

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				stream?.Dispose();
			}
			stream = null;

			base.Dispose(disposing);
		}

		public SFtpFileSystemPersister(string path, int flags) : base(path, flags)
		{
			FileMode mode;
			FileAccess access;

			if ((flags & SFtpFileMode.SSH_FXF_WRITE) != 0)
			{
				if (((flags & (SFtpFileMode.SSH_FXF_CREAT | SFtpFileMode.SSH_FXF_EXCL)) == (SFtpFileMode.SSH_FXF_CREAT | SFtpFileMode.SSH_FXF_EXCL))
					&& File.Exists(path))
				{
					throw new SFtpServerError(SFtpServerError.SFtpFileExistsError, SFtpStatus.SSH_FX_FAILURE);
				}

				if ((flags & (SFtpFileMode.SSH_FXF_CREAT | SFtpFileMode.SSH_FXF_TRUNC)) == (SFtpFileMode.SSH_FXF_CREAT | SFtpFileMode.SSH_FXF_TRUNC))
				{
					mode = FileMode.Create;
					access = FileAccess.Write;
				}
				else if (((flags & SFtpFileMode.SSH_FXF_CREAT) != 0) && !File.Exists(path))
				{
					mode = FileMode.Create;
					access = FileAccess.Write;
				}
				else if ((flags & SFtpFileMode.SSH_FXF_READ) != 0)
				{
					mode = FileMode.Open;
					access = FileAccess.ReadWrite;
				}
				else
				{
					mode = FileMode.Open;
					access = FileAccess.Write;
				}

				Directory.CreateDirectory(FileUtils.ExtractFilePath(path));
			}
			else if ((flags & SFtpFileMode.SSH_FXF_READ) != 0)
			{
				mode = FileMode.Open;
				access = FileAccess.Read;
			}
			else
			{
				throw new SFtpServerError(SFtpServerError.SFtpInvalidParameters, SFtpStatus.SSH_FX_FAILURE);
			}
			appendMode = ((flags & SFtpFileMode.SSH_FXF_APPEND) != 0);

			stream = new FileStream(path, mode, access);
		}

		public override int Read(byte[] buffer, int bufOffset, long offset, int count)
		{
			stream.Position = offset;
			return stream.Read(buffer, bufOffset, count);
		}

		public override int Write(byte[] buffer, int bufOffset, long offset, int count)
		{
			if (appendMode)
			{
				stream.Seek(0, SeekOrigin.End);
			}
			else
			{
				stream.Position = offset;
			}

			stream.Write(buffer, bufOffset, count);
			return count;
		}

		public override void GetAttributes(SFtpFileAttrs fileAttrs)
		{
			fileAttrs.SetSize(stream.Length);

			FileInfo fi = new FileInfo(Path);

			fileAttrs.SetDate(fi.LastAccessTimeUtc, fi.LastWriteTimeUtc);

			if (fi.IsReadOnly)
			{
				fileAttrs.SetPermissions(SFtpFilePermissions.OwnerRead | SFtpFilePermissions.GroupRead | SFtpFilePermissions.OtherRead);
			}
			else
			{
				fileAttrs.SetPermissions(
					SFtpFilePermissions.OwnerRead | SFtpFilePermissions.OwnerWrite | SFtpFilePermissions.OwnerExec |
					SFtpFilePermissions.GroupRead | SFtpFilePermissions.GroupWrite | SFtpFilePermissions.GroupExec |
					SFtpFilePermissions.OtherRead | SFtpFilePermissions.OtherWrite | SFtpFilePermissions.OtherExec
				);
			}
		}

		public override void SetAttributes(SFtpFileAttrs fileAttrs) { }
	}
}
